/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2011 Klaus Spanderen

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.

 compile:
g++ -O3 -I/home/spanderen/workspace/QuantLib -I/usr/local/include/boost-1_38 -o fdmscheme fdmscheme.cpp -L/home/spanderen/workspace/QuantLib/ql/.libs -lQuantLib

*/

#include <ql/quotes/simplequote.hpp>
#include <ql/time/calendars/target.hpp>
#include <ql/time/calendars/nullcalendar.hpp>
#include <ql/time/daycounters/actual365fixed.hpp>
#include <ql/instruments/europeanoption.hpp>
#include <ql/processes/hestonprocess.hpp>
#include <ql/processes/hullwhiteprocess.hpp>
#include <ql/processes/blackscholesprocess.hpp>
#include <ql/termstructures/yield/flatforward.hpp>
#include <ql/termstructures/volatility/equityfx/blackconstantvol.hpp>
#include <ql/processes/hybridhestonhullwhiteprocess.hpp>
#include <ql/math/optimization/levenbergmarquardt.hpp>
#include <ql/math/interpolations/cubicinterpolation.hpp>
#include <ql/models/equity/hestonmodelhelper.hpp>
#include <ql/pricingengines/vanilla/analytichestonengine.hpp>
#include <ql/pricingengines/vanilla/analyticbsmhullwhiteengine.hpp>
#include <ql/experimental/finitedifferences/fdmbackwardsolver.hpp>
#include <ql/experimental/finitedifferences/fdhestonvanillaengine.hpp>
#include <ql/experimental/finitedifferences/fdmhestonvariancemesher.hpp>
#include <ql/experimental/finitedifferences/fdblackscholesvanillaengine.hpp>
#include <ql/experimental/finitedifferences/fdhestonhullwhitevanillaengine.hpp>

#include <iostream>

#define LENGTH(x) (sizeof(x)/sizeof(x[0]))

using namespace QuantLib;

namespace {

    struct HestonModelData {
        const char* const name;
        Real v0;
        Real kappa;
        Real theta;
        Real sigma;
        Real rho;
        Real r;
        Real q;
    };
    
    HestonModelData hestonModels[] = {
        // ADI finite difference schemes for option pricing in the 
        // Heston model with correlation, K.J. in t'Hout and S. Foulon,
        {"'t Hout case 1", 0.04, 1.5, 0.04, 0.3, -0.9, 0.025, 0.0},
        {"'t Hout case 2", 0.12, 3.0, 0.12, 0.04, 0.6, 0.01, 0.04},
        {"'t Hout case 3", 0.0707,0.6067, 0.0707, 0.2928, -0.7571, 0.03, 0.0},
        {"'t Hout case 4", 0.06, 2.5, 0.06, 0.5, -0.1, 0.0507, 0.0469},
        // Efficient numerical methods for pricing American options under 
        // stochastic volatility, Samuli Ikonen and Jari Toivanen,
        {"Ikonen-Toivanen", 0.0625, 5, 0.16, 0.9, 0.1, 0.1, 0.0},
        // Not-so-complex logarithms in the Heston model, 
        // Christian Kahl and Peter Jäckel
        {"Kahl-Jaeckel", 0.16, 1.0, 0.16, 2.0, -0.8, 0.0, 0.0},
        // self defined test cases
        {"Equity case", 0.07, 2.0, 0.04, 0.55, -0.8, 0.03, 0.035 },
        {"high correlation", 0.07, 1.0, 0.04, 0.55,  0.9, 0.02, 0.04 },
        {"low Vol-Of-Vol", 0.07, 1.0, 0.04, 0.001, -0.75, 0.04, 0.03 },
        {"kappaEqSigRho", 0.07, 0.4, 0.04, 0.5, 0.8, 0.03, 0.03 }
    };

    struct HullWhiteModelData {
        const char* const name;
        Real a;
        Real sigma;
    };
    
    HullWhiteModelData hullWhiteModels[] = {
        {"EUR-2003", 0.00883, 0.00631 }
    };
    
    
    struct SchemeData {
        const char* const name;
        FdmSchemeDesc schemeDesc;
    };
    
    SchemeData schemes[] = {
        { "HV2", FdmSchemeDesc::ModifiedHundsdorfer() },
        { "HV",  FdmSchemeDesc::Hundsdorfer() },
        { "CS" , FdmSchemeDesc::CraigSneyd() },
        { "MCS", FdmSchemeDesc::ModifiedCraigSneyd() },
        { "DS" , FdmSchemeDesc::Douglas() }
    };

    struct VanillaOptionData {
        Real strike;
        Time maturity;
        Option::Type optionType;
    };

    struct FdHestonHullWhiteGrid {
        const char* const name;
        Size timeSteps, xGrid, vGrid, rGrid, dampingSteps;
    };

    boost::shared_ptr<YieldTermStructure>
    flatRate(Rate forward, const DayCounter& dc) {
        return boost::shared_ptr<YieldTermStructure>(
            new FlatForward(0, NullCalendar(), 
                 Handle<Quote>(boost::shared_ptr<SimpleQuote>(
                     new SimpleQuote(forward))), dc));
    }

    boost::shared_ptr<BlackVolTermStructure>
    flatVol(Volatility vol, const DayCounter& dc) {
        return boost::shared_ptr<BlackVolTermStructure>(
            new BlackConstantVol(0, NullCalendar(), 
                 Handle<Quote>(boost::shared_ptr<SimpleQuote>(
                     new SimpleQuote(vol))), dc));
    }

    boost::shared_ptr<HestonProcess> makeHestonProcess(
                                             const HestonModelData& params) {

        Handle<Quote> spot(
                boost::shared_ptr<SimpleQuote>(new SimpleQuote(100)));

        DayCounter dayCounter = Actual365Fixed();
        Handle<YieldTermStructure> rTS(flatRate(params.r, dayCounter));
        Handle<YieldTermStructure> qTS(flatRate(params.q, dayCounter));
        
        return boost::shared_ptr<HestonProcess>(new HestonProcess(
                   rTS, qTS, spot, params.v0, params.kappa, 
                   params.theta, params.sigma, params.rho));
    }
    
    boost::shared_ptr<VanillaOption> makeVanillaOption(
                                            const VanillaOptionData& params) {
        
        Date maturity = Date(Settings::instance().evaluationDate()) 
                                      + Period(Size(params.maturity*365), Days);
        boost::shared_ptr<Exercise> exercise(new EuropeanExercise(maturity));
        boost::shared_ptr<StrikedTypePayoff> payoff(
                    new PlainVanillaPayoff(params.optionType, params.strike));
        
        return boost::shared_ptr<VanillaOption>(
                                          new VanillaOption(payoff, exercise));
    }
}

void testHestonHullWhiteControlVariate(bool controlVariate) {

    std::cout << "Heston Hull-White" << std::endl;
    Date today(27, December, 2004);
    Settings::instance().evaluationDate() = today;
    Real equityIrCorr = -0.4;
    
	const Time maturity = 5.0;
    const Real strikes[] = {75,85,90,95,100,105,110,115,120,125,130,140,150};

    FdHestonHullWhiteGrid grids[] = { { "5x20x10x5",     5,  20,  10,  5, 0 },
									  { "10x20x10x10",  10,  20,  10, 10, 0 },
									  { "10x50x10x10",  10,  50,  10, 10, 0 },
									  { "10x50x20x10",  10,  50,  20, 10, 0 },
									  { "20x50x20x10",  20,  50,  20, 10, 0 },
									  { "20x50x20x20",  20,  50,  20, 20, 0 },
									  { "20x100x50x10", 20, 100,  50, 10, 0 },
									  { "20x100x50x20", 20, 100,  50, 20, 0 },
									  { "20x200x50x10", 20, 200,  50, 10, 0 },
									  { "20x200x100x10",20, 200, 100, 10, 0 },
									  { "20x100x100x20",20, 100, 100, 20, 0 },
									  { "50x200x100x20",50, 200, 100, 20, 0 } };
    FdHestonHullWhiteGrid refGrid = { "reference", 50, 400, 100, 40, 0 };
    
    for (Size k=0; k < LENGTH(hestonModels); ++k) {
        boost::shared_ptr<HestonProcess> hestonProcess(
									 makeHestonProcess(hestonModels[k]));
        boost::shared_ptr<HestonModel> hestonModel(
											new HestonModel(hestonProcess));

        boost::shared_ptr<HullWhiteProcess> hwProcess(
                new HullWhiteProcess(hestonProcess->riskFreeRate(),
                                     hullWhiteModels[0].a,
                                     hullWhiteModels[0].sigma));
            
        boost::shared_ptr<FdHestonHullWhiteVanillaEngine> refEngine(
                new FdHestonHullWhiteVanillaEngine(
                    hestonModel, hwProcess, equityIrCorr,  
                    refGrid.timeSteps, refGrid.xGrid, 
                    refGrid.vGrid, refGrid.rGrid,
                    refGrid.dampingSteps, true,
                    FdmSchemeDesc::ModifiedCraigSneyd()));
        refEngine->enableMultipleStrikesCaching(
                std::vector<Real>(strikes, 
                                  strikes + LENGTH(strikes)));
            
        for (Size i=0; i < LENGTH(schemes); ++i) {
            SchemeData scheme = schemes[i];
            for (Size j=0; j < LENGTH(grids); ++j) {
                Real avgPriceDiff = 0;

                boost::shared_ptr<FdHestonHullWhiteVanillaEngine> fdEngine(
						new FdHestonHullWhiteVanillaEngine(
							hestonModel, hwProcess, equityIrCorr,  
							grids[j].timeSteps, grids[j].xGrid, 
							grids[j].vGrid, grids[j].rGrid,
							grids[j].dampingSteps,
							controlVariate,
							scheme.schemeDesc));
                fdEngine->enableMultipleStrikesCaching(
						std::vector<Real>(strikes, 
										  strikes + LENGTH(strikes)));

                for (Size l=0; l < LENGTH(strikes); ++l) {
                    VanillaOptionData optionData 
						     = { strikes[l], maturity, Option::Call }; 
                    boost::shared_ptr<VanillaOption> option 
											= makeVanillaOption(optionData);
                    option->setPricingEngine(refEngine);
                    Real expected = option->NPV();
	     
                    option->setPricingEngine(fdEngine);
                    Real calculated = option->NPV();
                    avgPriceDiff
							+=std::fabs(expected-calculated)/LENGTH(strikes);
                }
                std::cout << scheme.name << "\t" 
                          << hestonModels[k].name << "\t"
                          << grids[j].name << "\t"
                          << ((controlVariate) ? "CV: on" : "CV: off ") << "\t"
                          << avgPriceDiff << std::endl;
			}
        }
    }
}


int main() {
    testHestonHullWhiteControlVariate(true);
    testHestonHullWhiteControlVariate(false);
}

